/* Gpx Viewer 
 * Copyright (C) 2009-2009 Qball Cow <qball@sarine.nl>
 * Project homepage: http://blog.sarine.nl/ 

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <glib.h>
#include <glib-object.h>
#include <config.h>
#include <gtk/gtk.h>
#include <pango/pango.h>
#include <cairo.h>
#include <time.h>
#include <gdk/gdk.h>
#include <float.h>
#include <math.h>
#include <pango/pangocairo.h>
#include <stdlib.h>
#include <string.h>
#include <glib/gi18n-lib.h>


#define GPX_TYPE_GRAPH (gpx_graph_get_type ())
#define GPX_GRAPH(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), GPX_TYPE_GRAPH, GpxGraph))
#define GPX_GRAPH_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), GPX_TYPE_GRAPH, GpxGraphClass))
#define GPX_IS_GRAPH(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GPX_TYPE_GRAPH))
#define GPX_IS_GRAPH_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), GPX_TYPE_GRAPH))
#define GPX_GRAPH_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), GPX_TYPE_GRAPH, GpxGraphClass))

typedef struct _GpxGraph GpxGraph;
typedef struct _GpxGraphClass GpxGraphClass;
typedef struct _GpxGraphPrivate GpxGraphPrivate;

#define GPX_TYPE_TRACK (gpx_track_get_type ())
#define GPX_TRACK(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), GPX_TYPE_TRACK, GpxTrack))
#define GPX_TRACK_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), GPX_TYPE_TRACK, GpxTrackClass))
#define GPX_IS_TRACK(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GPX_TYPE_TRACK))
#define GPX_IS_TRACK_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), GPX_TYPE_TRACK))
#define GPX_TRACK_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), GPX_TYPE_TRACK, GpxTrackClass))

typedef struct _GpxTrack GpxTrack;
typedef struct _GpxTrackClass GpxTrackClass;

#define GPX_TYPE_POINT (gpx_point_get_type ())
#define GPX_POINT(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), GPX_TYPE_POINT, GpxPoint))
#define GPX_POINT_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), GPX_TYPE_POINT, GpxPointClass))
#define GPX_IS_POINT(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GPX_TYPE_POINT))
#define GPX_IS_POINT_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), GPX_TYPE_POINT))
#define GPX_POINT_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), GPX_TYPE_POINT, GpxPointClass))

typedef struct _GpxPoint GpxPoint;
typedef struct _GpxPointClass GpxPointClass;
typedef struct _GpxTrackPrivate GpxTrackPrivate;

struct _GpxGraph {
	GtkEventBox parent_instance;
	GpxGraphPrivate * priv;
	GpxTrack* track;
};

struct _GpxGraphClass {
	GtkEventBoxClass parent_class;
};

struct _GpxGraphPrivate {
	gint _smooth_factor;
	PangoFontDescription* fd;
	cairo_surface_t* surf;
	gint LEFT_OFFSET;
	gint BOTTOM_OFFSET;
	time_t highlight;
	GpxPoint* start;
	GpxPoint* stop;
};

/**
 * This class represents a Track in a gpx file.
 * The tracks contains the points connecting everything together. 
 * Info like total distance, average speed, moving speed/time etc are available.
 */
struct _GpxTrack {
	GObject parent_instance;
	GpxTrackPrivate * priv;
	double total_distance;
	double max_speed;
	GList* points;
	GpxPoint* top;
	GpxPoint* bottom;
};

struct _GpxTrackClass {
	GObjectClass parent_class;
};


static gpointer gpx_graph_parent_class = NULL;

#define LOG_DOMAIN "GPX_PARSER"
#define unique_graph VERSION
GType gpx_graph_get_type (void);
GType gpx_track_get_type (void);
gpointer gpx_point_ref (gpointer instance);
void gpx_point_unref (gpointer instance);
GParamSpec* gpx_param_spec_point (const gchar* name, const gchar* nick, const gchar* blurb, GType object_type, GParamFlags flags);
void gpx_value_set_point (GValue* value, gpointer v_object);
gpointer gpx_value_get_point (const GValue* value);
GType gpx_point_get_type (void);
#define GPX_GRAPH_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GPX_TYPE_GRAPH, GpxGraphPrivate))
enum  {
	GPX_GRAPH_DUMMY_PROPERTY,
	GPX_GRAPH_SMOOTH_FACTOR
};
void gpx_graph_set_highlight (GpxGraph* self, const time_t* highlight);
static void gpx_graph_size_allocate_cb (GpxGraph* self, const GdkRectangle* alloc);
static void _gpx_graph_size_allocate_cb_gtk_widget_size_allocate (GtkWidget* _sender, const GdkRectangle* allocation, gpointer self);
static gboolean gpx_graph_button_press_event_cb (GpxGraph* self, const GdkEventButton* event);
static gboolean _gpx_graph_button_press_event_cb_gtk_widget_button_press_event (GtkWidget* _sender, const GdkEventButton* event, gpointer self);
static gboolean gpx_graph_motion_notify_event_cb (GpxGraph* self, const GdkEventMotion* event);
static gboolean _gpx_graph_motion_notify_event_cb_gtk_widget_motion_notify_event (GtkWidget* _sender, const GdkEventMotion* event, gpointer self);
static gboolean gpx_graph_button_release_event_cb (GpxGraph* self, const GdkEventButton* event);
static gboolean _gpx_graph_button_release_event_cb_gtk_widget_button_release_event (GtkWidget* _sender, const GdkEventButton* event, gpointer self);
GpxGraph* gpx_graph_new (void);
GpxGraph* gpx_graph_construct (GType object_type);
void gpx_graph_set_track (GpxGraph* self, GpxTrack* track);
time_t gpx_track_get_total_time (GpxTrack* self);
time_t gpx_point_get_time (GpxPoint* self);
static GpxPoint* gpx_graph_get_point_from_position (GpxGraph* self, double x, double y);
static void gpx_graph_update_surface (GpxGraph* self, GtkWidget* win);
static gboolean gpx_graph_real_expose_event (GtkWidget* base, const GdkEventExpose* event);
gint gpx_graph_get_smooth_factor (GpxGraph* self);
double gpx_track_calculate_point_to_point_speed (GpxTrack* self, GpxPoint* a, GpxPoint* b);
double gpx_track_get_track_average (GpxTrack* self);
double gpx_track_calculate_moving_average (GpxTrack* self, GpxPoint* start, GpxPoint* stop, time_t* moving_time);
void gpx_graph_set_smooth_factor (GpxGraph* self, gint value);
static void gpx_graph_finalize (GObject* obj);
static void gpx_graph_get_property (GObject * object, guint property_id, GValue * value, GParamSpec * pspec);
static void gpx_graph_set_property (GObject * object, guint property_id, const GValue * value, GParamSpec * pspec);


static void g_cclosure_user_marshal_VOID__OBJECT_POINTER_POINTER (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data);

void gpx_graph_set_highlight (GpxGraph* self, const time_t* highlight) {
	g_return_if_fail (self != NULL);
	self->priv->highlight = (*highlight);
	/* Force a redraw */
	gtk_widget_queue_draw ((GtkWidget*) self);
}


static void _gpx_graph_size_allocate_cb_gtk_widget_size_allocate (GtkWidget* _sender, const GdkRectangle* allocation, gpointer self) {
	gpx_graph_size_allocate_cb (self, allocation);
}


static gboolean _gpx_graph_button_press_event_cb_gtk_widget_button_press_event (GtkWidget* _sender, const GdkEventButton* event, gpointer self) {
	return gpx_graph_button_press_event_cb (self, event);
}


static gboolean _gpx_graph_motion_notify_event_cb_gtk_widget_motion_notify_event (GtkWidget* _sender, const GdkEventMotion* event, gpointer self) {
	return gpx_graph_motion_notify_event_cb (self, event);
}


static gboolean _gpx_graph_button_release_event_cb_gtk_widget_button_release_event (GtkWidget* _sender, const GdkEventButton* event, gpointer self) {
	return gpx_graph_button_release_event_cb (self, event);
}


GpxGraph* gpx_graph_construct (GType object_type) {
	GpxGraph * self;
	PangoFontDescription* _tmp0_;
	self = g_object_newv (object_type, 0, NULL);
	_tmp0_ = NULL;
	self->priv->fd = (_tmp0_ = pango_font_description_new (), (self->priv->fd == NULL) ? NULL : (self->priv->fd = (pango_font_description_free (self->priv->fd), NULL)), _tmp0_);
	pango_font_description_set_family (self->priv->fd, "sans mono");
	g_object_set ((GtkWidget*) self, "app-paintable", TRUE, NULL);
	gtk_event_box_set_visible_window ((GtkEventBox*) self, TRUE);
	g_signal_connect_object ((GtkWidget*) self, "size-allocate", (GCallback) _gpx_graph_size_allocate_cb_gtk_widget_size_allocate, self, 0);
	g_signal_connect_object ((GtkWidget*) self, "button-press-event", (GCallback) _gpx_graph_button_press_event_cb_gtk_widget_button_press_event, self, 0);
	g_signal_connect_object ((GtkWidget*) self, "motion-notify-event", (GCallback) _gpx_graph_motion_notify_event_cb_gtk_widget_motion_notify_event, self, 0);
	g_signal_connect_object ((GtkWidget*) self, "button-release-event", (GCallback) _gpx_graph_button_release_event_cb_gtk_widget_button_release_event, self, 0);
	return self;
}


GpxGraph* gpx_graph_new (void) {
	return gpx_graph_construct (GPX_TYPE_GRAPH);
}


void gpx_graph_set_track (GpxGraph* self, GpxTrack* track) {
	GpxTrack* _tmp1_;
	GpxTrack* _tmp0_;
	cairo_surface_t* _tmp2_;
	GpxPoint* _tmp3_;
	GpxPoint* _tmp4_;
	gboolean _tmp5_;
	g_return_if_fail (self != NULL);
	self->priv->highlight = 0;
	_tmp1_ = NULL;
	_tmp0_ = NULL;
	self->track = (_tmp1_ = (_tmp0_ = track, (_tmp0_ == NULL) ? NULL : g_object_ref (_tmp0_)), (self->track == NULL) ? NULL : (self->track = (g_object_unref (self->track), NULL)), _tmp1_);
	/* Invalidate the previous plot, so it is redrawn */
	_tmp2_ = NULL;
	self->priv->surf = (_tmp2_ = NULL, (self->priv->surf == NULL) ? NULL : (self->priv->surf = (cairo_surface_destroy (self->priv->surf), NULL)), _tmp2_);
	/* Force a redraw */
	gtk_widget_queue_draw ((GtkWidget*) self);
	_tmp3_ = NULL;
	self->priv->start = (_tmp3_ = NULL, (self->priv->start == NULL) ? NULL : (self->priv->start = (gpx_point_unref (self->priv->start), NULL)), _tmp3_);
	_tmp4_ = NULL;
	self->priv->stop = (_tmp4_ = NULL, (self->priv->stop == NULL) ? NULL : (self->priv->stop = (gpx_point_unref (self->priv->stop), NULL)), _tmp4_);
	_tmp5_ = FALSE;
	if (self->track != NULL) {
		_tmp5_ = self->track->points != NULL;
	} else {
		_tmp5_ = FALSE;
	}
	/* */
	if (_tmp5_) {
		g_signal_emit_by_name (self, "selection-changed", self->track, (GpxPoint*) g_list_first (self->track->points)->data, (GpxPoint*) g_list_last (self->track->points)->data);
	} else {
		g_signal_emit_by_name (self, "selection-changed", self->track, NULL, NULL);
	}
}


/**
 * Private functions
 */
static GpxPoint* gpx_graph_get_point_from_position (GpxGraph* self, double x, double y) {
	GpxPoint* result;
	gboolean _tmp0_;
	g_return_val_if_fail (self != NULL, NULL);
	if (self->track == NULL) {
		result = NULL;
		return result;
	}
	_tmp0_ = FALSE;
	if (x > self->priv->LEFT_OFFSET) {
		_tmp0_ = x < (((GtkWidget*) self)->allocation.width - 10);
	} else {
		_tmp0_ = FALSE;
	}
	if (_tmp0_) {
		double elapsed_time;
		time_t time;
		GList* iter;
		elapsed_time = (double) gpx_track_get_total_time (self->track);
		time = (time_t) (((x - self->priv->LEFT_OFFSET) / ((((GtkWidget*) self)->allocation.width - 10) - self->priv->LEFT_OFFSET)) * elapsed_time);
		iter = g_list_first (self->track->points);
		/* calculated time is offset from start time,  get real time */
		time = time + gpx_point_get_time ((GpxPoint*) iter->data);
		while (TRUE) {
			gboolean _tmp1_;
			if (!(iter->next != NULL)) {
				break;
			}
			_tmp1_ = FALSE;
			if (time < gpx_point_get_time ((GpxPoint*) iter->next->data)) {
				gboolean _tmp2_;
				_tmp2_ = FALSE;
				if (time == gpx_point_get_time ((GpxPoint*) iter->data)) {
					_tmp2_ = TRUE;
				} else {
					_tmp2_ = time > gpx_point_get_time ((GpxPoint*) iter->data);
				}
				_tmp1_ = _tmp2_;
			} else {
				_tmp1_ = FALSE;
			}
			if (_tmp1_) {
				GpxPoint* _tmp3_;
				_tmp3_ = NULL;
				result = (_tmp3_ = (GpxPoint*) iter->data, (_tmp3_ == NULL) ? NULL : gpx_point_ref (_tmp3_));
				return result;
			}
			iter = iter->next;
		}
	}
	result = NULL;
	return result;
}


static gboolean gpx_graph_button_press_event_cb (GpxGraph* self, const GdkEventButton* event) {
	gboolean result;
	GpxPoint* point;
	g_return_val_if_fail (self != NULL, FALSE);
	if (self->track == NULL) {
		result = FALSE;
		return result;
	}
	point = gpx_graph_get_point_from_position (self, (*event).x, (*event).y);
	if (point != NULL) {
		if ((*event).button == 1) {
			GpxPoint* _tmp1_;
			GpxPoint* _tmp0_;
			_tmp1_ = NULL;
			_tmp0_ = NULL;
			self->priv->start = (_tmp1_ = (_tmp0_ = point, (_tmp0_ == NULL) ? NULL : gpx_point_ref (_tmp0_)), (self->priv->start == NULL) ? NULL : (self->priv->start = (gpx_point_unref (self->priv->start), NULL)), _tmp1_);
		} else {
			GpxPoint* _tmp2_;
			_tmp2_ = NULL;
			self->priv->start = (_tmp2_ = NULL, (self->priv->start == NULL) ? NULL : (self->priv->start = (gpx_point_unref (self->priv->start), NULL)), _tmp2_);
			g_signal_emit_by_name (self, "point-clicked", point);
		}
	}
	result = FALSE;
	return result;
}


static gboolean gpx_graph_motion_notify_event_cb (GpxGraph* self, const GdkEventMotion* event) {
	gboolean result;
	GpxPoint* point;
	g_return_val_if_fail (self != NULL, FALSE);
	if (self->track == NULL) {
		result = FALSE;
		return result;
	}
	if (self->priv->start == NULL) {
		result = FALSE;
		return result;
	}
	point = gpx_graph_get_point_from_position (self, (*event).x, (*event).y);
	if (point != NULL) {
		GpxPoint* _tmp1_;
		GpxPoint* _tmp0_;
		gboolean _tmp2_;
		_tmp1_ = NULL;
		_tmp0_ = NULL;
		self->priv->stop = (_tmp1_ = (_tmp0_ = point, (_tmp0_ == NULL) ? NULL : gpx_point_ref (_tmp0_)), (self->priv->stop == NULL) ? NULL : (self->priv->stop = (gpx_point_unref (self->priv->stop), NULL)), _tmp1_);
		/* queue redraw so the selection is updated */
		gtk_widget_queue_draw ((GtkWidget*) self);
		_tmp2_ = FALSE;
		if (self->priv->start != NULL) {
			_tmp2_ = self->priv->stop != NULL;
		} else {
			_tmp2_ = FALSE;
		}
		if (_tmp2_) {
			if (gpx_point_get_time (self->priv->start) != gpx_point_get_time (self->priv->stop)) {
				if (gpx_point_get_time (self->priv->start) < gpx_point_get_time (self->priv->stop)) {
					g_signal_emit_by_name (self, "selection-changed", self->track, self->priv->start, self->priv->stop);
				} else {
					g_signal_emit_by_name (self, "selection-changed", self->track, self->priv->stop, self->priv->start);
				}
				result = FALSE;
				return result;
			}
		}
		g_signal_emit_by_name (self, "selection-changed", self->track, (GpxPoint*) g_list_first (self->track->points)->data, (GpxPoint*) g_list_last (self->track->points)->data);
	}
	result = FALSE;
	return result;
}


static gboolean gpx_graph_button_release_event_cb (GpxGraph* self, const GdkEventButton* event) {
	gboolean result;
	GpxPoint* point;
	g_return_val_if_fail (self != NULL, FALSE);
	if (self->track == NULL) {
		result = FALSE;
		return result;
	}
	point = gpx_graph_get_point_from_position (self, (*event).x, (*event).y);
	if (point != NULL) {
		if ((*event).button == 1) {
			GpxPoint* _tmp1_;
			GpxPoint* _tmp0_;
			_tmp1_ = NULL;
			_tmp0_ = NULL;
			self->priv->stop = (_tmp1_ = (_tmp0_ = point, (_tmp0_ == NULL) ? NULL : gpx_point_ref (_tmp0_)), (self->priv->stop == NULL) ? NULL : (self->priv->stop = (gpx_point_unref (self->priv->stop), NULL)), _tmp1_);
		} else {
			GpxPoint* _tmp2_;
			_tmp2_ = NULL;
			self->priv->stop = (_tmp2_ = NULL, (self->priv->stop == NULL) ? NULL : (self->priv->stop = (gpx_point_unref (self->priv->stop), NULL)), _tmp2_);
		}
		gtk_widget_queue_draw ((GtkWidget*) self);
		if ((*event).button == 1) {
			gboolean _tmp3_;
			_tmp3_ = FALSE;
			if (self->priv->start != NULL) {
				_tmp3_ = self->priv->stop != NULL;
			} else {
				_tmp3_ = FALSE;
			}
			if (_tmp3_) {
				if (gpx_point_get_time (self->priv->start) != gpx_point_get_time (self->priv->stop)) {
					if (gpx_point_get_time (self->priv->start) < gpx_point_get_time (self->priv->stop)) {
						g_signal_emit_by_name (self, "selection-changed", self->track, self->priv->start, self->priv->stop);
					} else {
						g_signal_emit_by_name (self, "selection-changed", self->track, self->priv->stop, self->priv->start);
					}
					result = FALSE;
					return result;
				}
			}
			g_signal_emit_by_name (self, "selection-changed", self->track, (GpxPoint*) g_list_first (self->track->points)->data, (GpxPoint*) g_list_last (self->track->points)->data);
		}
	}
	result = FALSE;
	return result;
}


static void gpx_graph_size_allocate_cb (GpxGraph* self, const GdkRectangle* alloc) {
	cairo_surface_t* _tmp0_;
	g_return_if_fail (self != NULL);
	/* Invalidate the previous plot, so it is redrawn */
	_tmp0_ = NULL;
	self->priv->surf = (_tmp0_ = NULL, (self->priv->surf == NULL) ? NULL : (self->priv->surf = (cairo_surface_destroy (self->priv->surf), NULL)), _tmp0_);
}


static gboolean gpx_graph_real_expose_event (GtkWidget* base, const GdkEventExpose* event) {
	GpxGraph * self;
	gboolean result;
	cairo_t* ctx;
	gboolean _tmp1_;
	self = (GpxGraph*) base;
	ctx = gdk_cairo_create ((GdkDrawable*) ((GtkWidget*) self)->window);
	/* If no valid surface, render it */
	if (self->priv->surf == NULL) {
		gpx_graph_update_surface (self, (GtkWidget*) self);
	}
	/* Draw the actual surface on the widget */
	cairo_set_source_surface (ctx, self->priv->surf, (double) 0, (double) 0);
	gdk_cairo_region (ctx, (*event).region);
	cairo_clip (ctx);
	cairo_paint (ctx);
	if (self->priv->highlight > 0) {
		GpxPoint* _tmp0_;
		GpxPoint* f;
		double elapsed_time;
		double graph_width;
		double graph_height;
		double hl;
		_tmp0_ = NULL;
		f = (_tmp0_ = (GpxPoint*) g_list_first (self->track->points)->data, (_tmp0_ == NULL) ? NULL : gpx_point_ref (_tmp0_));
		elapsed_time = (double) gpx_track_get_total_time (self->track);
		graph_width = (double) ((((GtkWidget*) self)->allocation.width - self->priv->LEFT_OFFSET) - 10);
		graph_height = (double) ((((GtkWidget*) self)->allocation.height - 20) - self->priv->BOTTOM_OFFSET);
		hl = ((self->priv->highlight - gpx_point_get_time (f)) / elapsed_time) * graph_width;
		cairo_translate (ctx, (double) self->priv->LEFT_OFFSET, (double) 20);
		cairo_set_source_rgba (ctx, 0.8, 0.2, 0.3, 0.8);
		cairo_move_to (ctx, hl, (double) 0);
		cairo_line_to (ctx, hl, graph_height);
		cairo_stroke_preserve (ctx);
		cairo_fill (ctx);
		(f == NULL) ? NULL : (f = (gpx_point_unref (f), NULL));
	}
	_tmp1_ = FALSE;
	if (self->priv->start != NULL) {
		_tmp1_ = self->priv->stop != NULL;
	} else {
		_tmp1_ = FALSE;
	}
	/* Draw selection, if available */
	if (_tmp1_) {
		if (gpx_point_get_time (self->priv->start) != gpx_point_get_time (self->priv->stop)) {
			GpxPoint* _tmp2_;
			GpxPoint* f;
			double elapsed_time;
			double graph_width;
			double graph_height;
			_tmp2_ = NULL;
			f = (_tmp2_ = (GpxPoint*) g_list_first (self->track->points)->data, (_tmp2_ == NULL) ? NULL : gpx_point_ref (_tmp2_));
			elapsed_time = (double) gpx_track_get_total_time (self->track);
			graph_width = (double) ((((GtkWidget*) self)->allocation.width - self->priv->LEFT_OFFSET) - 10);
			graph_height = (double) ((((GtkWidget*) self)->allocation.height - 20) - self->priv->BOTTOM_OFFSET);
			cairo_translate (ctx, (double) self->priv->LEFT_OFFSET, (double) 20);
			cairo_set_source_rgba (ctx, 0.3, 0.2, 0.3, 0.8);
			cairo_rectangle (ctx, ((gpx_point_get_time (self->priv->start) - gpx_point_get_time (f)) / elapsed_time) * graph_width, (double) 0, ((gpx_point_get_time (self->priv->stop) - gpx_point_get_time (self->priv->start)) / elapsed_time) * graph_width, graph_height);
			cairo_stroke_preserve (ctx);
			cairo_fill (ctx);
			(f == NULL) ? NULL : (f = (gpx_point_unref (f), NULL));
		}
	}
	result = FALSE;
	(ctx == NULL) ? NULL : (ctx = (cairo_destroy (ctx), NULL));
	return result;
}


static glong string_get_length (const char* self) {
	glong result;
	g_return_val_if_fail (self != NULL, 0L);
	result = g_utf8_strlen (self, -1);
	return result;
}


static void gpx_graph_update_surface (GpxGraph* self, GtkWidget* win) {
	cairo_t* ctx;
	cairo_surface_t* _tmp0_;
	cairo_t* _tmp1_;
	double max_speed;
	double elapsed_time;
	GpxPoint* _tmp5_;
	GpxPoint* f;
	double graph_width;
	double graph_height;
	PangoLayout* layout;
	double j;
	double step_size;
	char* _tmp7_;
	double _tmp8_;
	double size;
	GList* iter;
	guint interval;
	gint current;
	guint i;
	double avg;
	time_t moving_time = {0};
	gint w;
	gint h;
	g_return_if_fail (self != NULL);
	g_return_if_fail (win != NULL);
	ctx = gdk_cairo_create ((GdkDrawable*) win->window);
	_tmp0_ = NULL;
	self->priv->surf = (_tmp0_ = cairo_surface_create_similar (cairo_get_target (ctx), CAIRO_CONTENT_COLOR_ALPHA, win->allocation.width, win->allocation.height), (self->priv->surf == NULL) ? NULL : (self->priv->surf = (cairo_surface_destroy (self->priv->surf), NULL)), _tmp0_);
	_tmp1_ = NULL;
	ctx = (_tmp1_ = cairo_create (self->priv->surf), (ctx == NULL) ? NULL : (ctx = (cairo_destroy (ctx), NULL)), _tmp1_);
	g_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "gpx-graph.vala:253: Updating surface");
	/* Paint background white */
	cairo_set_source_rgba (ctx, (double) 1, (double) 1, (double) 1, (double) 1);
	cairo_paint (ctx);
	if (self->track == NULL) {
		(ctx == NULL) ? NULL : (ctx = (cairo_destroy (ctx), NULL));
		return;
	}
	max_speed = (double) 0;
	if (gpx_graph_get_smooth_factor (self) != 1) {
		GList* iter;
		iter = g_list_first (self->track->points);
		while (TRUE) {
			GList* ii;
			double speed;
			gint i;
			gint sf;
			double _tmp4_;
			if (!(iter->next != NULL)) {
				break;
			}
			ii = iter->next;
			speed = (double) 0;
			i = 0;
			sf = gpx_graph_get_smooth_factor (self);
			{
				gboolean _tmp2_;
				i = 0;
				_tmp2_ = TRUE;
				while (TRUE) {
					gboolean _tmp3_;
					if (!_tmp2_) {
						i++;
					}
					_tmp2_ = FALSE;
					_tmp3_ = FALSE;
					if (i < sf) {
						_tmp3_ = ii->prev != NULL;
					} else {
						_tmp3_ = FALSE;
					}
					if (!_tmp3_) {
						break;
					}
					speed = speed + gpx_track_calculate_point_to_point_speed (self->track, (GpxPoint*) ii->prev->data, (GpxPoint*) ii->data);
					ii = ii->prev;
				}
			}
			speed = speed / i;
			_tmp4_ = 0.0;
			if (speed > max_speed) {
				_tmp4_ = speed;
			} else {
				_tmp4_ = max_speed;
			}
			max_speed = _tmp4_;
			iter = iter->next;
		}
	} else {
		max_speed = self->track->max_speed;
	}
	elapsed_time = (double) gpx_track_get_total_time (self->track);
	g_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "gpx-graph.vala:283: Max speed: %f, elapsed_time: %f", max_speed, elapsed_time);
	cairo_translate (ctx, (double) self->priv->LEFT_OFFSET, (double) 20);
	_tmp5_ = NULL;
	f = (_tmp5_ = (GpxPoint*) self->track->points->data, (_tmp5_ == NULL) ? NULL : gpx_point_ref (_tmp5_));
	/* Draw Grid */
	graph_width = (double) ((win->allocation.width - self->priv->LEFT_OFFSET) - 10);
	graph_height = (double) ((win->allocation.height - 20) - self->priv->BOTTOM_OFFSET);
	layout = pango_cairo_create_layout (ctx);
	j = 0.0;
	step_size = graph_height / 8.0;
	cairo_set_source_rgba (ctx, 0.2, 0.2, 0.2, 0.6);
	cairo_set_line_width (ctx, (double) 1);
	{
		gboolean _tmp6_;
		j = graph_height;
		_tmp6_ = TRUE;
		while (TRUE) {
			if (!_tmp6_) {
				j = j - step_size;
			}
			_tmp6_ = FALSE;
			if (!(j > 0.0)) {
				break;
			}
			cairo_move_to (ctx, 0.0, j);
			cairo_line_to (ctx, graph_width, j);
			cairo_stroke (ctx);
		}
	}
	g_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "gpx-graph.vala:304: Draw grid lines");
	/* Draw speed and ticks */
	cairo_set_source_rgba (ctx, 0.0, 0.0, 0.0, 1.0);
	_tmp7_ = NULL;
	size = (_tmp8_ = (double) (self->priv->LEFT_OFFSET / string_get_length (_tmp7_ = g_strdup_printf ("%.1f", max_speed))), _tmp7_ = (g_free (_tmp7_), NULL), _tmp8_);
	if (size > step_size) {
		size = step_size;
	}
	pango_font_description_set_absolute_size (self->priv->fd, size * 1024);
	pango_layout_set_font_description (layout, self->priv->fd);
	{
		gboolean _tmp9_;
		j = (double) 0;
		_tmp9_ = TRUE;
		while (TRUE) {
			double speed;
			char* text;
			gint w;
			gint h;
			if (!_tmp9_) {
				j = j + step_size;
			}
			_tmp9_ = FALSE;
			if (!(j < graph_height)) {
				break;
			}
			speed = max_speed * ((graph_height - j) / graph_height);
			text = g_strdup_printf ("%.1f", speed);
			w = 0;
			h = 0;
			pango_layout_set_text (layout, text, -1);
			pango_layout_get_pixel_size (layout, &w, &h);
			cairo_move_to (ctx, (double) ((-w) - 5), j - (h / 2.0));
			pango_cairo_layout_path (ctx, layout);
			cairo_fill (ctx);
			g_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "gpx-graph.vala:321: Set speed tick: %s", text);
			cairo_move_to (ctx, (double) (-4), j);
			cairo_line_to (ctx, (double) 0, j);
			cairo_stroke (ctx);
			text = (g_free (text), NULL);
		}
	}
	/* Draw axis */
	cairo_move_to (ctx, 0.0, 0.0);
	cairo_set_source_rgba (ctx, 0.0, 0.0, 0.0, (double) 1);
	cairo_set_line_width (ctx, 1.5);
	cairo_line_to (ctx, 0.0, graph_height);
	cairo_line_to (ctx, graph_width, graph_height);
	cairo_stroke (ctx);
	g_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "gpx-graph.vala:337: Draw Axis");
	/* Draw the graph */
	cairo_set_source_rgba (ctx, 0.1, 0.2, 0.3, (double) 1);
	cairo_set_line_width (ctx, (double) 1);
	iter = g_list_first (self->track->points);
	cairo_move_to (ctx, 0.0, graph_height);
	while (TRUE) {
		double time_offset;
		double speed;
		GList* ii;
		gint i;
		gint sf;
		if (!(iter->next != NULL)) {
			break;
		}
		time_offset = (double) (gpx_point_get_time ((GpxPoint*) iter->data) - gpx_point_get_time (f));
		speed = (double) 0;
		ii = iter->next;
		i = 0;
		sf = gpx_graph_get_smooth_factor (self);
		{
			gboolean _tmp10_;
			i = 0;
			_tmp10_ = TRUE;
			while (TRUE) {
				gboolean _tmp11_;
				if (!_tmp10_) {
					i++;
				}
				_tmp10_ = FALSE;
				_tmp11_ = FALSE;
				if (i < sf) {
					_tmp11_ = ii->prev != NULL;
				} else {
					_tmp11_ = FALSE;
				}
				if (!_tmp11_) {
					break;
				}
				speed = speed + gpx_track_calculate_point_to_point_speed (self->track, (GpxPoint*) ii->prev->data, (GpxPoint*) ii->data);
				ii = ii->prev;
			}
		}
		speed = speed / i;
		cairo_line_to (ctx, graph_width * ((double) (time_offset / ((double) elapsed_time))), graph_height * ((double) (1.0 - (speed / max_speed))));
		iter = iter->next;
	}
	cairo_line_to (ctx, graph_width, graph_height);
	cairo_close_path (ctx);
	cairo_stroke_preserve (ctx);
	cairo_set_source_rgba (ctx, 0.1, 0.2, 0.8, 0.5);
	cairo_fill (ctx);
	g_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "gpx-graph.vala:368: Draw graph");
	iter = g_list_first (self->track->points);
	cairo_set_source_rgba (ctx, 0.0, 0.0, 0.0, 1.0);
	pango_font_description_set_absolute_size (self->priv->fd, (double) (12 * 1024));
	pango_layout_set_font_description (layout, self->priv->fd);
	interval = ((guint) elapsed_time) / ((guint) (graph_width / (5 * 12.0)));
	current = 0;
	i = 0U;
	{
		gboolean _tmp12_;
		i = (guint) 0;
		_tmp12_ = TRUE;
		while (TRUE) {
			gboolean _tmp13_;
			if (!_tmp12_) {
				i = i + interval;
			}
			_tmp12_ = FALSE;
			_tmp13_ = FALSE;
			if (i < elapsed_time) {
				_tmp13_ = interval > 0;
			} else {
				_tmp13_ = FALSE;
			}
			if (!_tmp13_) {
				break;
			}
			if ((graph_width * (1 - (i / elapsed_time))) > (2.5 * 12)) {
				gint w;
				gint h;
				char* text;
				w = 0;
				h = 0;
				text = g_strdup_printf ("%02i:%02i", ((gint) i) / 3600, (((gint) i) % 3600) / 60);
				pango_layout_set_text (layout, text, -1);
				pango_layout_get_pixel_size (layout, &w, &h);
				cairo_move_to (ctx, (graph_width * ((double) (i / elapsed_time))) - (w / 2.0), graph_height + 10);
				pango_cairo_layout_path (ctx, layout);
				cairo_fill (ctx);
				g_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "gpx-graph.vala:389: Set time tick: %s", text);
				cairo_move_to (ctx, graph_width * ((double) (i / elapsed_time)), graph_height);
				cairo_line_to (ctx, graph_width * ((double) (i / elapsed_time)), graph_height + 5);
				cairo_stroke (ctx);
				text = (g_free (text), NULL);
			}
			current++;
		}
	}
	/* Draw average speed */
	avg = gpx_track_get_track_average (self->track);
	cairo_set_source_rgba (ctx, 0.0, 0.7, 0.0, 0.7);
	cairo_move_to (ctx, 0.0, graph_height * (1 - (avg / max_speed)));
	cairo_line_to (ctx, graph_width, graph_height * (1 - (avg / max_speed)));
	cairo_stroke (ctx);
	g_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "gpx-graph.vala:405: Draw average speed line @ %.02f km/h", avg);
	/* Draw moving speed */
	avg = gpx_track_calculate_moving_average (self->track, (GpxPoint*) g_list_first (self->track->points)->data, (GpxPoint*) g_list_last (self->track->points)->data, &moving_time);
	cairo_set_source_rgba (ctx, 0.7, 0.0, 0.0, 0.7);
	cairo_move_to (ctx, 0.0, graph_height * (1 - (avg / max_speed)));
	cairo_line_to (ctx, graph_width, graph_height * (1 - (avg / max_speed)));
	cairo_stroke (ctx);
	g_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "gpx-graph.vala:415: Draw moving average speed line @ %.02f km/h", avg);
	/* Draw the title */
	w = 0;
	h = 0;
	cairo_set_source_rgba (ctx, 0.0, 0.0, 0.0, 1.0);
	pango_font_description_set_absolute_size (self->priv->fd, (double) (12 * 1024));
	pango_layout_set_font_description (layout, self->priv->fd);
	if (gpx_graph_get_smooth_factor (self) != 1) {
		char* markup;
		markup = g_strdup_printf (_ ("Speed (km/h) vs Time (HH:MM) <i>(smooth window: %i)</i>"), gpx_graph_get_smooth_factor (self));
		pango_layout_set_markup (layout, markup, -1);
		g_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "gpx-graph.vala:426: Set graph title: %s", markup);
		markup = (g_free (markup), NULL);
	} else {
		const char* _tmp14_;
		char* text;
		_tmp14_ = NULL;
		text = (_tmp14_ = _ ("Speed (km/h) vs Time (HH:MM)"), (_tmp14_ == NULL) ? NULL : g_strdup (_tmp14_));
		pango_layout_set_text (layout, text, -1);
		g_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "gpx-graph.vala:433: Set graph title: %s", text);
		text = (g_free (text), NULL);
	}
	pango_layout_get_pixel_size (layout, &w, &h);
	cairo_move_to (ctx, (graph_width / 2) - (w / 2), (double) (-20));
	pango_cairo_layout_path (ctx, layout);
	cairo_fill (ctx);
	(ctx == NULL) ? NULL : (ctx = (cairo_destroy (ctx), NULL));
	(f == NULL) ? NULL : (f = (gpx_point_unref (f), NULL));
	(layout == NULL) ? NULL : (layout = (g_object_unref (layout), NULL));
}


gint gpx_graph_get_smooth_factor (GpxGraph* self) {
	gint result;
	g_return_val_if_fail (self != NULL, 0);
	result = self->priv->_smooth_factor;
	return result;
}


void gpx_graph_set_smooth_factor (GpxGraph* self, gint value) {
	cairo_surface_t* _tmp0_;
	g_return_if_fail (self != NULL);
	self->priv->_smooth_factor = value;
	/* Invalidate the previous plot, so it is redrawn */
	_tmp0_ = NULL;
	self->priv->surf = (_tmp0_ = NULL, (self->priv->surf == NULL) ? NULL : (self->priv->surf = (cairo_surface_destroy (self->priv->surf), NULL)), _tmp0_);
	/* Force a redraw */
	gtk_widget_queue_draw ((GtkWidget*) self);
	g_object_notify ((GObject *) self, "smooth-factor");
}


static void gpx_graph_class_init (GpxGraphClass * klass) {
	gpx_graph_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (GpxGraphPrivate));
	GTK_WIDGET_CLASS (klass)->expose_event = gpx_graph_real_expose_event;
	G_OBJECT_CLASS (klass)->get_property = gpx_graph_get_property;
	G_OBJECT_CLASS (klass)->set_property = gpx_graph_set_property;
	G_OBJECT_CLASS (klass)->finalize = gpx_graph_finalize;
	g_object_class_install_property (G_OBJECT_CLASS (klass), GPX_GRAPH_SMOOTH_FACTOR, g_param_spec_int ("smooth-factor", "smooth-factor", "smooth-factor", G_MININT, G_MAXINT, 4, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_signal_new ("point_clicked", GPX_TYPE_GRAPH, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__POINTER, G_TYPE_NONE, 1, GPX_TYPE_POINT);
	g_signal_new ("selection_changed", GPX_TYPE_GRAPH, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_user_marshal_VOID__OBJECT_POINTER_POINTER, G_TYPE_NONE, 3, GPX_TYPE_TRACK, GPX_TYPE_POINT, GPX_TYPE_POINT);
}


static void gpx_graph_instance_init (GpxGraph * self) {
	self->priv = GPX_GRAPH_GET_PRIVATE (self);
	self->priv->_smooth_factor = 4;
	self->track = NULL;
	self->priv->fd = NULL;
	self->priv->surf = NULL;
	self->priv->LEFT_OFFSET = 60;
	self->priv->BOTTOM_OFFSET = 30;
	self->priv->highlight = 0;
	self->priv->start = NULL;
	self->priv->stop = NULL;
}


static void gpx_graph_finalize (GObject* obj) {
	GpxGraph * self;
	self = GPX_GRAPH (obj);
	(self->track == NULL) ? NULL : (self->track = (g_object_unref (self->track), NULL));
	(self->priv->fd == NULL) ? NULL : (self->priv->fd = (pango_font_description_free (self->priv->fd), NULL));
	(self->priv->surf == NULL) ? NULL : (self->priv->surf = (cairo_surface_destroy (self->priv->surf), NULL));
	(self->priv->start == NULL) ? NULL : (self->priv->start = (gpx_point_unref (self->priv->start), NULL));
	(self->priv->stop == NULL) ? NULL : (self->priv->stop = (gpx_point_unref (self->priv->stop), NULL));
	G_OBJECT_CLASS (gpx_graph_parent_class)->finalize (obj);
}


GType gpx_graph_get_type (void) {
	static GType gpx_graph_type_id = 0;
	if (gpx_graph_type_id == 0) {
		static const GTypeInfo g_define_type_info = { sizeof (GpxGraphClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) gpx_graph_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (GpxGraph), 0, (GInstanceInitFunc) gpx_graph_instance_init, NULL };
		gpx_graph_type_id = g_type_register_static (GTK_TYPE_EVENT_BOX, "GpxGraph", &g_define_type_info, 0);
	}
	return gpx_graph_type_id;
}


static void gpx_graph_get_property (GObject * object, guint property_id, GValue * value, GParamSpec * pspec) {
	GpxGraph * self;
	gpointer boxed;
	self = GPX_GRAPH (object);
	switch (property_id) {
		case GPX_GRAPH_SMOOTH_FACTOR:
		g_value_set_int (value, gpx_graph_get_smooth_factor (self));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void gpx_graph_set_property (GObject * object, guint property_id, const GValue * value, GParamSpec * pspec) {
	GpxGraph * self;
	self = GPX_GRAPH (object);
	switch (property_id) {
		case GPX_GRAPH_SMOOTH_FACTOR:
		gpx_graph_set_smooth_factor (self, g_value_get_int (value));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}



static void g_cclosure_user_marshal_VOID__OBJECT_POINTER_POINTER (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data) {
	typedef void (*GMarshalFunc_VOID__OBJECT_POINTER_POINTER) (gpointer data1, gpointer arg_1, gpointer arg_2, gpointer arg_3, gpointer data2);
	register GMarshalFunc_VOID__OBJECT_POINTER_POINTER callback;
	register GCClosure * cc;
	register gpointer data1, data2;
	cc = (GCClosure *) closure;
	g_return_if_fail (n_param_values == 4);
	if (G_CCLOSURE_SWAP_DATA (closure)) {
		data1 = closure->data;
		data2 = param_values->data[0].v_pointer;
	} else {
		data1 = param_values->data[0].v_pointer;
		data2 = closure->data;
	}
	callback = (GMarshalFunc_VOID__OBJECT_POINTER_POINTER) (marshal_data ? marshal_data : cc->callback);
	callback (data1, g_value_get_object (param_values + 1), gpx_value_get_point (param_values + 2), gpx_value_get_point (param_values + 3), data2);
}



