/* Gpx Viewer 
 * Copyright (C) 2009-2009 Qball Cow <qball@sarine.nl>
 * Project homepage: http://blog.sarine.nl/ 

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <glib.h>
#include <glib-object.h>
#include <config.h>
#include <time.h>
#include <float.h>
#include <math.h>


#define GPX_TYPE_PLAYBACK (gpx_playback_get_type ())
#define GPX_PLAYBACK(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), GPX_TYPE_PLAYBACK, GpxPlayback))
#define GPX_PLAYBACK_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), GPX_TYPE_PLAYBACK, GpxPlaybackClass))
#define GPX_IS_PLAYBACK(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GPX_TYPE_PLAYBACK))
#define GPX_IS_PLAYBACK_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), GPX_TYPE_PLAYBACK))
#define GPX_PLAYBACK_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), GPX_TYPE_PLAYBACK, GpxPlaybackClass))

typedef struct _GpxPlayback GpxPlayback;
typedef struct _GpxPlaybackClass GpxPlaybackClass;
typedef struct _GpxPlaybackPrivate GpxPlaybackPrivate;

#define GPX_TYPE_TRACK (gpx_track_get_type ())
#define GPX_TRACK(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), GPX_TYPE_TRACK, GpxTrack))
#define GPX_TRACK_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), GPX_TYPE_TRACK, GpxTrackClass))
#define GPX_IS_TRACK(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GPX_TYPE_TRACK))
#define GPX_IS_TRACK_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), GPX_TYPE_TRACK))
#define GPX_TRACK_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), GPX_TYPE_TRACK, GpxTrackClass))

typedef struct _GpxTrack GpxTrack;
typedef struct _GpxTrackClass GpxTrackClass;

#define GPX_TYPE_POINT (gpx_point_get_type ())
#define GPX_POINT(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), GPX_TYPE_POINT, GpxPoint))
#define GPX_POINT_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), GPX_TYPE_POINT, GpxPointClass))
#define GPX_IS_POINT(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GPX_TYPE_POINT))
#define GPX_IS_POINT_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), GPX_TYPE_POINT))
#define GPX_POINT_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), GPX_TYPE_POINT, GpxPointClass))

typedef struct _GpxPoint GpxPoint;
typedef struct _GpxPointClass GpxPointClass;
typedef struct _GpxTrackPrivate GpxTrackPrivate;

struct _GpxPlayback {
	GObject parent_instance;
	GpxPlaybackPrivate * priv;
};

struct _GpxPlaybackClass {
	GObjectClass parent_class;
};

struct _GpxPlaybackPrivate {
	GpxTrack* track;
	guint timer;
	time_t progress;
	GList* current;
	GpxPoint* first;
};

/**
 * This class represents a Track in a gpx file.
 * The tracks contains the points connecting everything together. 
 * Info like total distance, average speed, moving speed/time etc are available.
 */
struct _GpxTrack {
	GObject parent_instance;
	GpxTrackPrivate * priv;
	double total_distance;
	double max_speed;
	GList* points;
	GpxPoint* top;
	GpxPoint* bottom;
};

struct _GpxTrackClass {
	GObjectClass parent_class;
};


static gpointer gpx_playback_parent_class = NULL;

#define GT_LOG_DOMAIN "GPX_PLAYBACK"
#define gt_unique_graph VERSION
GType gpx_playback_get_type (void);
GType gpx_track_get_type (void);
gpointer gpx_point_ref (gpointer instance);
void gpx_point_unref (gpointer instance);
GParamSpec* gpx_param_spec_point (const gchar* name, const gchar* nick, const gchar* blurb, GType object_type, GParamFlags flags);
void gpx_value_set_point (GValue* value, gpointer v_object);
gpointer gpx_value_get_point (const GValue* value);
GType gpx_point_get_type (void);
#define GPX_PLAYBACK_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GPX_TYPE_PLAYBACK, GpxPlaybackPrivate))
enum  {
	GPX_PLAYBACK_DUMMY_PROPERTY
};
GpxPlayback* gpx_playback_new (GpxTrack* track);
GpxPlayback* gpx_playback_construct (GType object_type, GpxTrack* track);
time_t gpx_point_get_time (GpxPoint* self);
gboolean gpx_playback_timer_callback (GpxPlayback* self);
void gpx_playback_stop (GpxPlayback* self);
static gboolean _gpx_playback_timer_callback_gsource_func (gpointer self);
void gpx_playback_start (GpxPlayback* self);
void gpx_playback_pause (GpxPlayback* self);
static void gpx_playback_finalize (GObject* obj);



GpxPlayback* gpx_playback_construct (GType object_type, GpxTrack* track) {
	GpxPlayback * self;
	GpxTrack* _tmp1_;
	GpxTrack* _tmp0_;
	g_return_val_if_fail (track != NULL, NULL);
	self = g_object_newv (object_type, 0, NULL);
	_tmp1_ = NULL;
	_tmp0_ = NULL;
	self->priv->track = (_tmp1_ = (_tmp0_ = track, (_tmp0_ == NULL) ? NULL : g_object_ref (_tmp0_)), (self->priv->track == NULL) ? NULL : (self->priv->track = (g_object_unref (self->priv->track), NULL)), _tmp1_);
	if (self->priv->track->points != NULL) {
		GpxPoint* _tmp3_;
		GpxPoint* _tmp2_;
		_tmp3_ = NULL;
		_tmp2_ = NULL;
		self->priv->first = (_tmp3_ = (_tmp2_ = (GpxPoint*) g_list_first (self->priv->track->points)->data, (_tmp2_ == NULL) ? NULL : gpx_point_ref (_tmp2_)), (self->priv->first == NULL) ? NULL : (self->priv->first = (gpx_point_unref (self->priv->first), NULL)), _tmp3_);
	}
	return self;
}


GpxPlayback* gpx_playback_new (GpxTrack* track) {
	return gpx_playback_construct (GPX_TYPE_PLAYBACK, track);
}


gboolean gpx_playback_timer_callback (GpxPlayback* self) {
	gboolean result;
	g_return_val_if_fail (self != NULL, FALSE);
	self->priv->progress = self->priv->progress + 30;
	g_debug ("gpx-playback.vala:50: Tick\n");
	while (TRUE) {
		gboolean _tmp0_;
		_tmp0_ = FALSE;
		if (self->priv->current != NULL) {
			_tmp0_ = gpx_point_get_time ((GpxPoint*) self->priv->current->data) < self->priv->progress;
		} else {
			_tmp0_ = FALSE;
		}
		if (!_tmp0_) {
			break;
		}
		self->priv->current = self->priv->current->next;
	}
	if (self->priv->current == NULL) {
		self->priv->timer = (guint) 0;
		g_signal_emit_by_name (self, "tick", NULL);
		g_debug ("gpx-playback.vala:55: stopping\n");
		result = FALSE;
		return result;
	}
	/* make larger steps */
	g_signal_emit_by_name (self, "tick", (GpxPoint*) self->priv->current->data);
	self->priv->current = self->priv->current->next;
	result = TRUE;
	return result;
}


static gboolean _gpx_playback_timer_callback_gsource_func (gpointer self) {
	return gpx_playback_timer_callback (self);
}


void gpx_playback_start (GpxPlayback* self) {
	g_return_if_fail (self != NULL);
	gpx_playback_stop (self);
	if (self->priv->first != NULL) {
		self->priv->progress = gpx_point_get_time (self->priv->first);
		g_debug ("gpx-playback.vala:69: start playback\n");
		self->priv->current = g_list_first (self->priv->track->points);
		self->priv->timer = g_timeout_add ((guint) 250, _gpx_playback_timer_callback_gsource_func, self);
	}
}


void gpx_playback_pause (GpxPlayback* self) {
	g_return_if_fail (self != NULL);
	if (self->priv->current == NULL) {
		return;
	}
	if (self->priv->timer > 0) {
		g_source_remove (self->priv->timer);
		self->priv->timer = (guint) 0;
	} else {
		self->priv->timer = g_timeout_add ((guint) 250, _gpx_playback_timer_callback_gsource_func, self);
	}
}


void gpx_playback_stop (GpxPlayback* self) {
	g_return_if_fail (self != NULL);
	if (self->priv->timer > 0) {
		g_debug ("gpx-playback.vala:87: stop playback\n");
		g_source_remove (self->priv->timer);
		self->priv->timer = (guint) 0;
	}
	g_signal_emit_by_name (self, "tick", NULL);
}


static void gpx_playback_class_init (GpxPlaybackClass * klass) {
	gpx_playback_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (GpxPlaybackPrivate));
	G_OBJECT_CLASS (klass)->finalize = gpx_playback_finalize;
	g_signal_new ("tick", GPX_TYPE_PLAYBACK, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__POINTER, G_TYPE_NONE, 1, GPX_TYPE_POINT);
}


static void gpx_playback_instance_init (GpxPlayback * self) {
	self->priv = GPX_PLAYBACK_GET_PRIVATE (self);
	self->priv->track = NULL;
	self->priv->timer = (guint) 0;
	self->priv->progress = 0;
	self->priv->first = NULL;
}


static void gpx_playback_finalize (GObject* obj) {
	GpxPlayback * self;
	self = GPX_PLAYBACK (obj);
	(self->priv->track == NULL) ? NULL : (self->priv->track = (g_object_unref (self->priv->track), NULL));
	(self->priv->first == NULL) ? NULL : (self->priv->first = (gpx_point_unref (self->priv->first), NULL));
	G_OBJECT_CLASS (gpx_playback_parent_class)->finalize (obj);
}


GType gpx_playback_get_type (void) {
	static GType gpx_playback_type_id = 0;
	if (gpx_playback_type_id == 0) {
		static const GTypeInfo g_define_type_info = { sizeof (GpxPlaybackClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) gpx_playback_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (GpxPlayback), 0, (GInstanceInitFunc) gpx_playback_instance_init, NULL };
		gpx_playback_type_id = g_type_register_static (G_TYPE_OBJECT, "GpxPlayback", &g_define_type_info, 0);
	}
	return gpx_playback_type_id;
}




